"""
These bits are about classifying characters for purposes of running a finite
state machine, not for representing character classes in regular expressions.

At the moment, the only supported mechanisms are simple: Store the boundaries
between codepoint ranges with identical properties. To minimize the alphabet
is to assign equivalence classes to these ranges. A runtime, binary search
identifies the correct range, and then a list index operation identifies the
final character class.

As an aside: The present scanner table compaction method is probably not too
sensitive to the order of equivalence classes.
"""

from abc import ABC, abstractmethod
import bisect
from .interface import Classifier
from ..support import pretty


class SimpleClassifier(Classifier):
	def __init__(self, bounds):
		self.bounds = tuple(bounds)
	def cardinality(self): return 1+len(self.bounds)
	def classify(self, codepoint:int): return bisect.bisect_right(self.bounds, codepoint)
	def display(self): pretty.print_grid([['-', *self.bounds]])

class MetaClassifier(Classifier):
	def __init__(self, bounds, classes):
		self.bounds = tuple(bounds)
		self.classes = tuple(classes)
		assert len(self.bounds) + 1 == len(self.classes)
		assert set(self.classes) == set(range(self.cardinality()))
	def cardinality(self): return max(self.classes)+1
	def classify(self, codepoint): return self.classes[bisect.bisect_right(self.bounds, codepoint)]
	def display(self):
		bound_repr = ["'"+chr(b) if 32 < b < 127 else b for b in self.bounds]
		pretty.print_grid([('-', *bound_repr), self.classes, ])



"""
G-for-Generalized: GLR is basically the non-deterministic version of LR parsing.

As a parsing strategy, GLR means non-determinism is embraced and resolved as necessary at
in the context of actual parse runs. You can do GLR with any viable LR-style HFA, though
better tables naturally result in fewer blind alleys. In principle, the "trivial" table
is just the set of all grammar rules at every point: this yields classical Earley parsing.

Within this module, various handle-finding automaton (HFA) construction functions return a
(mostly) standardized structure which properly represents any remaining non-determinism.
Transmuting these to fully-deterministic structures is the LR module's job.

The HFA class itself contains a trial-parse routine which exercises these constructions
by determining whether they recognize a string as "in the language". It's great for testing
and illustrates one method to perform a parallel-parse, but it does not bother to recover a
parse tree or semantic value. Such things could be added, but preferably atop good means
for persisting ambiguous parse tables.
"""
import collections, sys
from pprint import pprint
from typing import Iterable, TypeVar, Generic, Any
from typing import NamedTuple
from ..support import foundation, pretty
from .interface import HandleFindingAutomaton, END_OF_TOKENS, ERROR_SYMBOL, ParseError
from .context_free import ContextFreeGrammar, Rule, SemanticAction, LEFT, RIGHT, NONASSOC, BOGUS

T = TypeVar('T')

class PurityError(ValueError):
	""" Raised if a grammar has the wrong/undeclared conflicts. """

class HFA(Generic[T]):
	"""
	HFA is short for "Handle-Finding Automaton", which is clunky to repeat all over the place.
	
	What is a handle?
	
	A handle is that point in the text where the right end of a rule has been matched and
	recognition of the corresponding non-terminal symbol could be performed.
	
	There is a common base structure for these, and certain operations are largely similar
	regardless of which sort of table-construction algorithm is underway.
	
	The main difference between HFAs is what sort of object they use for states.
	To participate in the generic operations, states must have a "shift" dictionary mapping
	symbols to successor state numbers (not objects) and support state.reductions_before(lexeme)
	which must return an Iterable of zero or more rule-IDs.
	
	Fields are:
	graph: a list of state objects; their index is implicitly their node ID.
	initial: a list of initial-state ID numbers (graph node indices) corresponding
		to the start-symbols of the CFG.
	accept: a list of final/accepting-state ID numbers (graph node indices) also
		corresponding to the start-symbols of the CFG.
	bft: The BreadthFirstTraversal object which was used for the construction.
		This happens to be greatly useful in various diagnostic and other capacities.
	"""
	graph: list[T]
	initial: list[int]
	accept: list[int]
	bft: foundation.BreadthFirstTraversal
	
	def __init__(self, *, graph, initial, accept, bft):
		""" I don't want exactly named-tuple semantics. """
		self.graph, self.initial,self.accept, self.bft = graph, initial, accept, bft
	
	def traverse(self, q: int, symbols:Iterable) -> int:
		""" Starting in state q, follow the shifts for symbols, and return the resulting state ID. """
		for s in symbols: q = self.graph[q].shift[s]
		return q
	
	def make_dot_file(self, path):
		""" Make a file suitable for the "dot" application from the Graphviz package. """
		# FIXME: This fails to take reductions into account.
		with open(path, 'w') as fh:
			fh.write("digraph {\n")
			for q, state in enumerate(self.graph):
				sym = self.bft.breadcrumbs[q] or ''
				sym = sym.replace('"', r'\"')
				if sym.endswith('\\'): sym = sym + ' '
				fh.write("%d [label=\"%d: %s\"]\n"%(q, q, sym))
				for i in state.shift.values():
					fh.write("\t%d -> %d\n"%(q,i))
			fh.write('}\n')
		pass
	
	def trial_parse(self, rules: list[Rule], sentence: Iterable[str]):
		"""
		This is intended to be a super-simplistic non-deterministic recognizer: It exists only for
		unit-testing the GLR table-construction algorithms, and therefore doesn't try to build a
		semantic value or worry about pathological grammars.

		The approach taken is a lock-step parallel simulation with a cactus-stack of viable states:
		each entry is a cons cell consisting of a state id and prior stack. Because it explores
		every possible parse, it will diverge if faced with an infinitely-ambiguous situation.
		There are ways to cope with such cases, but in practice they are normally the result of
		mistakes, so the more useful response is to reject infinitely-ambiguous grammars.

		To play along, HFA states must support the .reductions_before(lexeme) method.
		"""
		
		language_index = 0  # Or perhaps: grammar.start.index[language_symbol]
		initial, accept = self.initial[language_index], self.accept[language_index]
		
		def reduce(stack, rule_id):
			""" To perform a reduction, roll the stack to before the RHS and then shift the LHS. """
			rule = rules[rule_id]
			for i in range(len(rule.rhs)): stack = stack[1]
			return self.graph[stack[0]].shift[rule.lhs], stack
		
		root = (initial, None)
		alive = [root]
		for lexeme in sentence:
			next = []
			for stack in alive:
				state = self.graph[stack[0]]
				if lexeme in state.shift: next.append((state.shift[lexeme], stack))
				for rule_id in state.reductions_before(lexeme): alive.append(reduce(stack, rule_id))
			alive = next
			if not alive: raise ParseError("Parser died midway at something ungrammatical.")
		for stack in alive:
			q = stack[0]
			if q == accept: return True
			for rule_id in self.graph[q].reductions_before(END_OF_TOKENS):
				alive.append(reduce(stack, rule_id))
		raise ParseError("Parser recognized a viable prefix, but not a complete sentence.")


class LR0_State(NamedTuple):
	"""
	The LR(0) construction completely ignores right-context.
	Therefore, an LR(0) state tracks which rules it may recognize,
	but does not differentiate this information any further.
	"""
	shift: dict[str, int]  # symbol => state-id
	reduce: list[int]  # rule-id
	
	def reductions_before(self, lexeme):
		""" Did I mention LR(0) doesn't worry about look-ahead? """
		return self.reduce


class LookAheadState(NamedTuple):
	"""
	An LR(1) or LALR(1) table needs to take a token of right-context
	into account when recognizing a rule. Therefore, the .reduce field
	is a dictionary keyed by look-ahead token.
	"""
	shift: dict[str, int]  # symbol => state-id
	reduce: dict[str, list[int]]  # LALR
	
	def reductions_before(self, lexeme):
		return self.reduce.get(lexeme, ())



def reachable(step: dict, reduce: dict, grammar:ContextFreeGrammar) -> dict:
	"""
	This function exists so that parse-table construction algorithms can respect operator
	precedence and associativity declarations in a grammar specification. It is a vital
	part of minimal-LR1 mode, but it also works for LALR and canonical-LR1.
	
	The object of this function is to prevent the exploration of useless/unreachable states.
	It does this by deleting a shift from the "step" dictionary whenever said shift becomes
	impossible by virtue of P&A declarations. It also deletes useless reductions similarly
	rendered unavailable, and denotes non-associativity errors by a sentinel empty-tuple.

	NOTE: When shift/reduce/reduce conflicts arise in connection with operator-precedence
	specifications, the intended semantics for a generalized parse are not always clearly
	defined. The code below should either behave sensibly or toss an exception.

	The problem with the bizarre corner cases is they cannot be understood solely in terms
	of actions the parser might (or might not) take in this state. If they come up, you
	get a warning printed on STDERR but otherwise the operator-precedence declarations
	are ignored in that instance.
	
	FIXME: This function should not perform I/O directly.
	"""
	for token, rule_id_list in list(reduce.items()):
		if token not in step: continue
		decide = [grammar.decide_shift_reduce(token, rule_id) for rule_id in rule_id_list]
		ways = set(decide)
		assert BOGUS not in ways, "This is guaranteed by grammar.validate(...), called earlier."
		if len(ways) == 1:
			decision = ways.pop()
			if decision == LEFT:
				del step[token]
			elif decision == RIGHT:
				del reduce[token]
			elif decision == NONASSOC:
				del step[token]
				reduce[token] = ()
			else:
				assert decision is None
		elif ways == {LEFT, NONASSOC}:
			del step[token]
			reduce[token] = tuple(r for r, d in zip(rule_id_list, decide) if d == LEFT)
		elif ways == {RIGHT, None}:
			reduce[token] = tuple(r for r, d in zip(rule_id_list, decide) if d != RIGHT)
		else:
			print("Fair Warning:", token, "triggers a bizarre operator-precedence corner case.", file=sys.stderr)
	return step



class DragonBookTable(HandleFindingAutomaton):
	"""
	This is the classic textbook view of a set of parse tables: a pair of dense matrices
	(implemented here as lists-of-lists) representing the "ACTION" and "GOTO" tables, along
	with information about the reduction rules. The contents of these matrices are just
	numbers representing parse actions.
	
	This is a reasonable implementation as-is if you have a modern amount of RAM in your
	machine. In days of old, it would be necessary to compress the parse tables. Today,
	that's still not such a bad idea if you can pre-compute the tables. The compaction
	submodule contains some code for a typical method of parser table compression, and
	the runtime submodule implements the HandleFindingAutomaton interface atop a compressed table.
	
	Design note: There's a temptation to make the constructor take an HFA object, but
	that limits the ways you can instantiate this class. See the function `tabulate(...)`.
	"""
	
	def __init__(self, *, initial: dict, action: list, goto: list, nonassoc_errors: set, rules: list, terminals: list,
	             nonterminals: list, breadcrumbs: list, splits=()):
		self.initial = initial
		self.action_matrix = action
		self.goto_matrix = goto
		self.nonassoc_errors = nonassoc_errors
		self.translate = {symbol: i for i, symbol in enumerate(terminals)}
		nontranslate = {symbol: i for i, symbol in enumerate(nonterminals)}
		self.terminals, self.nonterminals = terminals, nonterminals
		self.breadcrumbs = breadcrumbs
		
		def translate_rule(rule:Rule):
			"""
			The current approach to storing rule actions is:
			Each rule gets a constructor number and a list of "places".
			A negative constructor-number means a bracketing-rule,
			with an offset from the stack pointer telling where to find the semantic item.
			A non-negative points into a table of semantic actions
			for use with a translated ``places`` vector, which gives argument offsets.
			Perhaps the ``places`` vector should also be part of the equivalence relationship?
			"""
			size = len(rule.rhs)
			if isinstance(rule.action, int):
				assert rule.action < size
				cid,places = rule.action-size, ()
			else:
				assert isinstance(rule.action, SemanticAction)
				cid,places = messages.classify(rule.action.message), tuple(x-size for x in rule.action.indices)
			return nontranslate[rule.lhs], size, cid, places
		
		messages = foundation.EquivalenceClassifier()
		self.rule_table = list(map(translate_rule, rules))
		self.constructors = messages.exemplars
		
		self.rule_provenance = [rule.provenance for rule in rules]
		self.splits = splits # A non-deterministic table just needs one extra bit: this list of lists.
		
		interactive = []
		for row in action:
			k = set(row)
			k.discard(0)
			if len(k) == 1: interactive.append(min(k.pop(), 0))
			else: interactive.append(0)
		for q, t in nonassoc_errors: interactive[q] = False
		self.interactive_step = interactive.__getitem__
	
	def get_rule(self, rule_id: int) -> tuple:
		return self.rule_table[rule_id]
	
	def get_translation(self, symbol) -> int:
		try: return self.translate[symbol]
		except KeyError: return len(self.terminals) # Guaranteed to trigger error-processing.
		
	def get_action(self, state_id, terminal_id) -> int:
		try: return self.action_matrix[state_id][terminal_id]
		except IndexError: return 0 # And this needs to not panic if a bad token is observed; just cry foul.
	
	def get_goto(self, state_id, nonterminal_id) -> int: return self.goto_matrix[state_id][nonterminal_id]
	
	def get_initial(self, language) -> int: return 0 if language is None else self.initial[language]
	
	def get_breadcrumb(self, state_id) -> str: return self.breadcrumbs[state_id]
	
	def display(self):
		size = len(self.action_matrix)
		print('Action and Goto: (%d states)' % size)
		head = ['', ''] + self.terminals + [''] + self.nonterminals
		body = []
		for i, (b, a, g) in enumerate(zip(self.breadcrumbs, self.action_matrix, self.goto_matrix)):
			body.append([i, b, *a, '', *g])
		pretty.print_grid([head] + body)
		if self.splits:
			print("Splits:")
			pprint(self.splits)
	
	def make_csv(self, pathstem):
		""" Generate action and goto tables into CSV files suitable for inspection in a spreadsheet program. """
		
		def mask(q, row, essential):
			return [
				s if s or (q, t) in essential else None
				for t, s in enumerate(row)
			]
		
		def typical_grid(top, matrix, essential):
			head = [None, None, *top]
			return [head] + [[q, self.breadcrumbs[q]] + mask(q, row, essential) for q, row in enumerate(matrix)]
		
		pretty.write_csv_grid(pathstem + '.action.csv',
			typical_grid(self.terminals, self.action_matrix, self.nonassoc_errors))
		pretty.write_csv_grid(pathstem + '.goto.csv', typical_grid(self.nonterminals, self.goto_matrix, frozenset()))
	
	def get_split_offset(self) -> int:
		return len(self.action_matrix)
	
	def get_split(self, split_id: int) -> list:
		assert split_id>0
		return self.splits[split_id]

	def get_constructor(self, constructor_id) -> Any: return self.constructors[constructor_id]
	
	def each_constructor(self):
		mentions = [set() for _ in self.constructors]
		for provenance, (ntid, size, cid, places) in zip(self.rule_provenance, self.rule_table):
			if cid >= 0:
				mentions[cid].add(provenance)
		return zip(self.constructors, mentions)


class ParsingStyle:
	"""
	There are three main ways to deal with inadequacies (non-determinism) remaining
	after application of any P&A declarations:
		1. Pure: Inadequacies are considered a grammar bug.
		2. Deterministic: Inadequacies are resolved to shift, or to use the earliest-defined rule.
		3. Generalized: Inadequacies are converted to parser-split entries.
	
	Probably the correct choice of style should be reflected in the grammar definition somehow.
	"""
	
	def decide_inadequacy(self, q:int, look_ahead:str, shift:int, rule_ids:Iterable, rules:list) -> int:
		""" Called in all non-deterministic situations. """
		raise NotImplementedError(type(self))
	
	def any_splits(self):
		""" Return nothing, or a list of splits for use in non-deterministic parsing algorithms. """
		raise NotImplementedError(type(self))

	def report(self, hfa):
		""" Give user-feedback about any observed challenges. """
		raise NotImplementedError(type(self))

class DeterministicStyle(ParsingStyle):
	
	def __init__(self, strict:bool):
		self.conflicts = collections.defaultdict(list)
		self.strict = strict
	
	def decide_inadequacy(self, q:int, look_ahead: str, shift: int, rule_ids: Iterable, rules:list) -> int:
		self.conflicts[q, look_ahead].extend(rule_ids)
		return shift or encode_reduce(min(rule_ids))
	
	def any_splits(self):
		pass
	
	def report(self, hfa):
		"""
		This function was originally intended as a way to visualize the branches of a conflict.
		In its original form a bunch of context was available; I've gratuitously stripped that away
		and now I want to break this down to the bits we actually need.
		
		BreadthFirstTraversal.traversal[x] was used to grab the core parse items in order to
		visualize the state reached by shifting the lookahead token if that shift is viable.
		Such really belongs as a method on the state: soon it will move there.
		
		The "options" list contains numeric candidate ACTION instructions which are interpreted
		in the usual way: This does represent a data-coupling, but one that's unlikely to change,
		so I'm not too worried about it just now.
		
		In conclusion: Let the objects defined in automata.py format parse-states for human consumption.
		"""
		if not self.conflicts:
			# print("Grammar specification is fully deterministic.")
			pass
		elif self.strict:
			raise PurityError(self.conflicts)
		else:
			print("Grammar specification contains conflicts.")
			print("Conflict reporting is presently undergoing an overhaul, but here's some diagnostic data:")
			pprint(self.conflicts)

class GeneralizedStyle(ParsingStyle):
	
	def __init__(self, splits_offset:int):
		self.offset = splits_offset
		self.splits = []
	
	def decide_inadequacy(self, q: int, look_ahead: str, shift: int, rule_ids: Iterable, rules:list) -> int:
		split = []
		if shift: split.append(shift)
		for r in sorted(rule_ids, key=lambda i:len(rules[i].rhs)): split.append(-1-r)
		return self.offset + foundation.allocate(self.splits, split)
	
	def any_splits(self):
		return self.splits
	
	def report(self, hfa):
		print(len(self.splits), "non-deterministic situation(s) encountered.")


def encode_reduce(rule_id:int) -> int:
	""" See interface.HandleFindingAutomaton.get_action. """
	return -1 - rule_id

def tabulate(hfa: HFA[LookAheadState], grammar:ContextFreeGrammar, *, style:ParsingStyle) -> DragonBookTable:
	"""
	Having an HFA based on State objects, this function produces a corresponding
	dense-matrix-style parse table of the sort typically shown in textbook descriptions
	of LR-style parsing automata.

	This function does NOT worry about precedence and associativity declarations:
	It assumes that concern has already been taken care of in the input HFA -
	principally by the interaction of function `reachable(...)` with the P&A bits.

	Any residual inadequacies of the grammar are delegated to the `style` object for
	resolution.
	"""
	assert isinstance(grammar, ContextFreeGrammar), grammar
	assert END_OF_TOKENS not in grammar.symbols
	assert ERROR_SYMBOL not in grammar.symbol_rule_ids
	terminals = [END_OF_TOKENS] + sorted(grammar.apparent_terminals())
	translate = {t:i for i,t in enumerate(terminals)}
	nonterminals = sorted(grammar.symbol_rule_ids.keys())
	##### Tabulate the states into dense matrices ACTION and GOTO:
	action, goto, nonassoc_errors = [], [], set()
	conflict = collections.defaultdict(set)
	for q, state in enumerate(hfa.graph):
		goto.append([state.shift.get(s, 0) for s in nonterminals])
		action_row = [state.shift.get(s, 0) for s in terminals]
		conflict.clear()
		for symbol, rule_ids in state.reduce.items():
			idx = translate[symbol]
			shift = action_row[idx]
			if rule_ids == ():
				# This is how function `reachable(...)` communicates a non-association situation.
				assert shift == 0
				nonassoc_errors.add((q,idx))
			elif shift == 0 and len(rule_ids) == 1:
				action_row[idx] = encode_reduce(rule_ids[0])
			else: action_row[idx] = style.decide_inadequacy(q, symbol, shift, rule_ids, grammar.rules)
		action.append(action_row)
	for q, t in nonassoc_errors: action[q][t] = 0
	for q in hfa.accept: action[q][0] = q
	style.report(hfa)
	return DragonBookTable(
		initial=dict(zip(grammar.start, hfa.initial)),
		action=action,
		goto=goto,
		nonassoc_errors=nonassoc_errors,
		rules=grammar.rules,
		terminals=terminals,
		nonterminals=nonterminals,
		breadcrumbs=hfa.bft.breadcrumbs,
		splits=style.any_splits()
	)





"""
Compile a macroparse grammar/scanner definition from a markdown document
into a set of parsing and scanning tables in JSON format.

The resulting tables are suitable for use with the included runtime modules.
  Pypi: https://pypi.org/project/booze-tools/
  GitHub: https://github.com/kjosib/booze-tools/wiki
  ReadTheDocs: https://boozetools.readthedocs.io/en/latest/
"""

import sys, os, argparse, json

from boozetools.macroparse.compiler import compile_string
from boozetools.macroparse.grammar import DefinitionError
from boozetools.macroparse import compaction

def parse_arguments():
	parser = argparse.ArgumentParser(prog='py -m boozetools', description=__doc__,)
	parser.add_argument('source_path', help='path to input file')
	parser.add_argument('-f', '--force', action='store_true', dest='force', help='allow to write over existing file')
	parser.add_argument('-o', '--output', help='path to output file')
	parser.add_argument('-i', '--indent', help='indent the JSON output for easier reading.', action='store_const', dest='indent', const=2, default=None)
	parser.add_argument('--pretty', action='store_true', help='Display uncompressed tables in attractive grid format on STDOUT.')
	parser.add_argument('--csv', action='store_true', help='Generate CSV versions of uncompressed tables, suitable for inspection.')
	parser.add_argument('--dev', action='store_true', help='Operate in "development mode" -- which changes from time to time.')
	parser.add_argument('--dot', action='store_true', help="Create a .dot file for visualizing the parser via the Graphviz package.")
	parser.add_argument('-v', '--verbose', action='store_true', help="Squawk, mainly about the table compression stats.")
	# if len(sys.argv) < 2: exit(parser.print_help())
	return parser.parse_args()

def main(args):
	if args.verbose: compaction.VERBOSE = True
	stem, extension = os.path.splitext(args.source_path)
	target_path = args.output or stem+'.automaton'
	if os.path.exists(target_path) and not args.force:
		print('Target file already exists and --force command-line argument was not given.', file=sys.stderr)
		exit(1)
	with(open(args.source_path)) as fh:document = fh.read()
	try:
		intermediate_form = compile_string(document)
		if args.dot: intermediate_form.make_dot_file(target_path+'.dot')
		textbook_form = intermediate_form.determinize()
	except DefinitionError as e:
		print(e.args[0], file=sys.stderr)
		exit(1)
	else:
		if args.pretty: textbook_form.pretty_print()
		if args.csv: textbook_form.make_csv(target_path)
		compact = textbook_form.as_compact_form(filename=os.path.basename(args.source_path))
		if args.dev:
			compact_goto = compact['parser']['goto']
			for listname in ['row_index', 'col_index', 'quotient']:
				print(listname+':', compact_goto[listname], len(compact_goto[listname]))
			print('mark:', compact_goto['mark'], 'residue:', len(compact_goto['quotient'])-compact_goto['mark'])
		json.dump(compact, open(target_path, 'w'), separators = (',', ':'), sort_keys = False, indent = args.indent)
		print('Wrote automaton in JSON format to:')
		print('\t'+target_path)

if __name__ == '__main__': main(parse_arguments())




"""
Scanning Interface Definitions.
"""
from abc import ABC, abstractmethod
from typing import TypeVar, NamedTuple, Optional, Any

State = TypeVar("State")
Symbol = int
RuleId = int

CodePoint = int
ConditionVector = tuple[State, State]

INITIAL = 'INITIAL'
END_OF_INPUT : CodePoint = -1

class ScannerBlocked(ValueError):
	"""
	Raised (by default) if a scanner gets blocked.
	Parameters are:
		the string offset where it happened.
		the current start-condition of the scanner.
	"""
	def __init__(self, position, condition):
		super().__init__(position, condition)
		self.position, self.condition = position, condition
		
		
class FiniteAutomaton(ABC):
	"""
	A finite automaton determines which rule matches but knows nothing about the rules themselves.
	This interface captures the operations required to execute the general scanning algorithm.
	It is deliberately decoupled from any particular representation of the underlying data.
	"""
	
	@abstractmethod
	def jam_state(self):
		""" DFA might provide -1, while NFA might provide an empty frozenset(). """

	@abstractmethod
	def condition(self, name:str) -> ConditionVector:
		""" A "condition" is (currently) a pair of state_ids for the normal and beginning-of-line cases. """
	
	@abstractmethod
	def transition(self, state: State, codepoint: CodePoint) -> State:
		""" The FSM's delta function. """
	
	@abstractmethod
	def accept(self, state: State) -> RuleId:
		""" Return the associated rule ID if this state is terminal, otherwise None. """
	
	
class Classifier(ABC):
	"""
	Normally a finite-state automaton (FA) based scanner does not represent all possible input
	characters as individual and distinct. Rather, all possible characters are mapped
	to a much smaller alphabet of symbols which are distinguishable from their neighbors
	in terms of their effect on the operation of the FA.
	It is this object's responsibility to perform that mapping via method `classify`.
	"""
	
	@abstractmethod
	def classify(self, codepoint: CodePoint) -> Symbol:
		"""
		Map a unicode codepoint to a specific numbered character class
		such that 0 <= result < self.cardinality()
		as known to a corresponding finite automaton.
		"""
	
	@abstractmethod
	def cardinality(self) -> int:
		""" Return the number of distinct classes which may be emitted by self.classify(...). """
	
	@abstractmethod
	def display(self):
		""" Pretty-print a suitable representation of the innards of this classifier's data. """

class Bindings(ABC):
	
	@abstractmethod
	def on_match(self, yy, rule_id:RuleId):
		""" Delegate. If there's right-context, call yy.less(amount) to implement it before delegating. """
		# act = self.__act[rule_id]
		# self.less(act.right_context)
		# self.__method[rule_id](self, *act.args)
		
	def on_stuck(self, yy):
		""" If you override this to return normally, scanning will continue normally afterward. """
		raise ScannerBlocked(yy.left, yy.condition)



"""
Time for a new tack.

A long LONG time ago, I built a parser generator in PHP.
It was partly based on a literal modeling of the ideas in [1],
but portions of that description were impenetrably vague.
I happened across [2] and did a full reverse-engineering job
to go from C code (with plenty of wacky pointer manipulation)
back to something conceptual, and finally understood what was going on.
So that became [3]. I stuck it on source-forge and forgot about it for years.
Eventually, I discovered Richard van Velzen had forked it,
fixed some upward compatibility issues, and put the result on GitHub.
So thank you Mr. van Velzen!

Over time, I wrote a few different LALR implementations.
Either I would misplace the source code or want to try out some new interface idea.
Each time, I tried to do the LALR generation part just a bit more nicely.
At one point, I fell into the trap known as NQLALR and a previous version of this code had that fault.
After finally (years later!) discovering my error, I set about trying to correct it.
Well, there's no easy way back. One must essentially start from scratch.

Bibliography:

1. Chapter 9 of Parsing Techniques: a Practical Guide, by Dick Grune and Ceriel Jacobs.
2. Source code to lemon.c parser generator. See [3] for one copy.
3. Source code to lime.php parser generator. https://github.com/rvanvelzen/lime

================================================================================
"""

from typing import NamedTuple
from ..support.foundation import allocate, strongly_connected_components_by_tarjan
from .interface import END_OF_TOKENS
from .context_free import ContextFreeGrammar
from .automata import HFA, LR0_State, LookAheadState, reachable
from .lr0 import lr0_construction, ParseItemMap


def lalr_construction(grammar: ContextFreeGrammar) -> HFA[LookAheadState]:
	"""
	Building a non-deterministic LALR(1)-style table is a direct extension of the LR(0)
	construction. LR(0) tables tend to have lots of inadequate states. If we figure out
	which look-ahead tokens are relevant to which reductions, then the automaton gets a
	good deal more capable. In the limit, there is canonical LR(1) as given by Knuth.
	That's traditionally been considered impractical for all but the very smallest
	grammars. Today's workstations have the chops to handle canonical LR(1) even for
	larger grammars, but it's still a chore and, as we will see, a needless one.

	The bulk of the work involved consists of finding the follow-sets for reductions.
	I've chosen to break that work into its own function, rather than code the entire
	LALR construction as one function, for two reasons. First, it's an interesting
	algorithm in its own right. Second, that function provides almost the perfect
	inputs for exploitation later in the minimal-LR(1) construction given later.

	If we stop here, we get a table that's more efficient than LR(0) for generalized
	parsing: it will attempt many fewer dead-end reductions before look-ahead tokens
	that LALR determines not to be in the reduction's follow-set. In fact, if you
	plan to use generalized parsing, LALR is probably your best choice to generate
	the parse table. A "stronger" table means more states to track and slow the parse.
	"""
	lr0 = lr0_construction(ParseItemMap.from_grammar(grammar))
	terminal_sets, reduce_set_id = find_lalr_sets(lr0, grammar)
	def make_lalr_state(q:int, node:LR0_State) -> LookAheadState:
		reduce = {}
		for rule_id in node.reduce:
			for terminal in terminal_sets[reduce_set_id[q, rule_id]]:
				if terminal not in reduce: reduce[terminal] = [rule_id]
				else: reduce[terminal].append(rule_id) # This branch represents an R/R conflict...
				
		# At this point, `LookAheadState(node.shift, reduce)` would be a nice, potentially
		# non-deterministic HFA state incorporating information about LALR follow sets.
		# We can do better: Calling upon the "reachable" function causes the resulting
		# non-deterministic HFA to respect precedence and associativity (P&A) declarations.
		# This parallels the corresponding call in the true-LR(1) algorithm, simplifies
		# the construction of a deterministic table, does not change the deterministic
		# semantics, and may have some subtle implications for GLR semantics in the presence
		# of P&A declarations. In particular, an invasive high-precedence reduction could
		# cause trouble. I've not taken the time to work out if that's even a thing, though.
		# For now, if you want to do GLR parsing with P&A, use the minimal_lr1(...) construction.
		
		step = reachable(node.shift, reduce, grammar)
		
		# That may leave unreachable states in the HFA graph, but a later pass can remove them.
		
		return LookAheadState(step, reduce)
	
	return HFA(
		graph=[make_lalr_state(q, node) for q, node in enumerate(lr0.graph)],
		initial=lr0.initial, accept=lr0.accept, bft=lr0.bft
	)
	
def find_lalr_sets(lr0:HFA[LR0_State], grammar:ContextFreeGrammar) -> tuple[list, dict[tuple[int,int], int]]:
	"""
	This is a variant of the channel algorithm as described (colorfully and frankly not
	very clearly) in chapter 9 of Parsing Techniques: a Practical Guide, by Dick Grune
	and Ceriel Jacobs. In essence the idea is to build a directed graph by certain rules
	and then "flow" terminal symbols around the nodes in that graph until updates cease.
	You may recognize this as a fix-point over the set-union operation. It turns out we
	can do better: using Tarjan's Strongly-Connected-Components algorithm and orienting
	the edges in the correct direction, we need only ever consider each edge once.

	:returns (terminal_sets, reduce_set_id) such that:
		1. terminal_sets[state_id] is equal to the "first" set of state_id
		2. terminal_sets[reduce_set_id[state_id, rule_id]] is the set of tokens which the LALR
		   algorithm determines can trigger reduction.
	"""
	graph = lr0.graph
	terminals = grammar.apparent_terminals()
	
	# Prepare to find first-sets:
	epsilon = grammar.find_epsilon()
	terminal_sets = [terminals.intersection(node.shift) for node in graph]
	for q in lr0.accept:terminal_sets[q].add(END_OF_TOKENS)
	inbound = [[dst for symbol, dst in node.shift.items() if symbol in epsilon] for node in graph]
	
	# Now do something about follow-sets and reduce-sets.
	
	# There is one follow-set for each nonterminal-edge in the graph.
	# There is one reduce-set for each reducing parse-item in a parse-item-core.
	# One must not confuse these two distinct ideas, lest one end up with NQ-LALR.
	nonterminals = frozenset(grammar.symbol_rule_ids)
	follow_set_id: dict[tuple[int, str], int] = {}
	reduce_set_id = {}
	for q_src, node in enumerate(graph):
		for rule_id in node.reduce:
			reduce_set_id[q_src, rule_id] = allocate(terminal_sets, set())
			inbound.append([])
		for symbol, q_dst in node.shift.items():
			if symbol in nonterminals:
				# A follow-set always contains the FIRST-set of the successor-state...
				follow_set_id[q_src, symbol] = allocate(terminal_sets, set())
				inbound.append([q_dst])
	
	# And, there is some kind of inclusion relation between them.
	prefix, suffix = prepare_rule_affixes(grammar.rules, epsilon, nonterminals)
	for (q_src, lhs), src_fs_id in follow_set_id.items():
		for rule_id in grammar.symbol_rule_ids[lhs]:
			# A prefix of the RHS symbols does not "see" this follow-set:
			q_mid = lr0.traverse(q_src, prefix[rule_id])
			# Suffix symbols can "see" this follow-set, so they must be connected up:
			for symbol in suffix[rule_id]:
				inbound[follow_set_id[q_mid, symbol]].append(src_fs_id)
				q_mid = graph[q_mid].shift[symbol]
			# The reduce-set is a union of follow-sets
			inbound[reduce_set_id[q_mid, rule_id]].append(src_fs_id)
			
	# Destructive-update to perform a union closure on terminal_sets.
	for component in strongly_connected_components_by_tarjan(inbound):
		union = set()
		for k in component:
			union.update(terminal_sets[k])
			union.update(*[terminal_sets[j] for j in inbound[k]])
			terminal_sets[k] = union
	
	return terminal_sets, reduce_set_id
	
def prepare_rule_affixes(rules, epsilon, nonterminals):
	"""
	The right-hand side of each rule can be divided into two parts, prefix and suffix.
	The suffix contains all those symbols which "see" the follow-set of the corresponding
	left-hand side as it sits in the grammar graph.
	"""
	# Incidentally, it seems likely that this same treatment could simplify some of the LR(1) code.
	# I'll have to think about it.
	def divide(rhs):
		if not rhs:
			return (), ()
		prefix, suffix = list(rhs), []
		while prefix and prefix[-1] in epsilon:
			suffix.insert(0, prefix.pop())
		if prefix and prefix[-1] in nonterminals:
			suffix.insert(0, prefix.pop())
		return prefix, suffix
	return zip(*[divide(rule.rhs) for rule in rules])




""" No frills. Plenty useful. """

import inspect
import sys

from . import automata, shift_reduce
from .interface import ParseErrorListener, SemanticError
from .context_free import Rule, SemanticAction, ContextFreeGrammar, LEFT, RIGHT, NONASSOC
from .all_methods import PARSE_TABLE_METHODS

class MiniParse(ParseErrorListener):
	""" Connects BNF production rules directly to Python functions. No frills. Very useful as-is. """
	def __init__(self, *start, method='LALR', strict=False):
		self.__grammar = ContextFreeGrammar()
		self.__grammar.start.extend(start)
		self.__attrribute = []
		self.__hfa: automata.DragonBookTable = None
		self.__combine = None
		self.__awaiting_action = False
		self.__method = method
		self.__strict = strict
		self.void_symbols = set() # These won't get picked up as arguments.
	
	def left(self, symbols:list): self.__grammar.assoc(LEFT, symbols)
	def right(self, symbols:list): self.__grammar.assoc(RIGHT, symbols)
	def nonassoc(self, symbols:list): self.__grammar.assoc(NONASSOC, symbols)
	
	def __analyze(self, rhs):
		rhs = rhs.split()
		args = []
		for i,symbol in enumerate(rhs):
			if symbol.startswith('.'):
				rhs[i] = symbol[1:]
				args.append(i)
		if not args: # Pick up non-void symbols if no dotted symbols appear in the right-hand side.
			args = tuple(i for i, symbol in enumerate(rhs) if symbol not in self.void_symbols)
		return tuple(rhs), args
	
	def rule(self, lhs:str, rhs_str:str, prec_sym=None):
		"""
		Decorates a callable as applying when the given rule is recognized.
		Prefix significant arguments with ., or get the whole right-hand side passed as arguments.
		For renaming and bracketing rules, call as
			foo.rule('expr', '( .expr )')(None)
		For normal rules, call as
			@foo.rule('expr', '.expr + .expr')
			def add(a,b): return a+b
		"""
		assert self.__hfa is None
		assert not self.__awaiting_action, "You forgot to provide the action for the prior production rule."
		self.__awaiting_action = True
		rhs, offsets = self.__analyze(rhs_str)
		def decorate(fn=None):
			hash(fn) # Because later it will need to be hashable; push failures to the fore.
			assert self.__awaiting_action
			self.__awaiting_action = False
			if fn is None:
				if len(rhs) == 1: action = 0 # Unit/renaming rule
				elif len(offsets) == 1: action = offsets[0]  # Bracketing rule
				else: action = SemanticAction(_collect_tuple, offsets)
			else: action = SemanticAction(fn, offsets)
			previous_frame = inspect.currentframe().f_back
			rule = Rule(lhs, rhs, prec_sym, action, inspect.getframeinfo(previous_frame)[:2])
			self.__grammar.add_rule(rule)
			return fn
		return decorate
	
	def renaming(self, lhs:str, *alternatives):
		"""
		Facilitates those "X => A | B | C | D" type rules you often find in real grammars.
		"""
		assert not self.__awaiting_action, "You forgot to provide the action for the prior production rule."
		for branch in alternatives:
			rhs, offsets = self.__analyze(branch)
			if len(rhs) == 1: action = 0  # Unit/renaming rule
			elif len(offsets) == 1: action = offsets[0]  # Bracketing rule
			else: raise AssertionError('%r is not a single-member branch -- although you could prepend the significant member with a dot ( like .this ) to fix it.' % branch)
			previous_frame = inspect.currentframe().f_back
			rule = Rule(lhs, rhs, None, action, inspect.getframeinfo(previous_frame)[:2])
			self.__grammar.add_rule(rule)
	
	def display(self): self.__grammar.display()

	def get_hfa_and_combine(self):
		if self.__hfa is None:
			if self.__awaiting_action: raise AssertionError('You forgot to provide the action for the final production rule.')
			self.__grammar.validate()
			self.__hfa = automata.tabulate(PARSE_TABLE_METHODS[self.__method](self.__grammar), self.__grammar,
										   style=automata.DeterministicStyle(self.__strict))
			constructors = self.__hfa.constructors
			def combine(constructor_id, args):
				try: return constructors[constructor_id](*args)
				except SemanticError as ex: raise ex from None
				except Exception as ex:
					message = self.__hfa.get_constructor(constructor_id)
					file_path = inspect.getfile(message)
					line_number = 1+inspect.findsource(message)[1]
					return self.exception_parsing(ex, file_path, line_number, args)
			self.__combine = combine
		return self.__hfa, self.__combine

	def parse(self, each_token, *, language=None):
		hfa, combine = self.get_hfa_and_combine()
		return shift_reduce.parse(hfa, combine, each_token, language=language, on_error=self)
	
	@staticmethod
	def exception_parsing(ex:Exception, file_path, line_number, args):
		print("\n---\nWhile trying to call %s:%d"%(file_path, line_number), file=sys.stderr)
		raise ex
	
def _collect_tuple(*items): return items



"""
Of all the LR-style parsing methods, LR(0) is the least sophisticated and the easiest to understand.
It's also the functional foundation for all the rest, so it makes sense to start reading here.
"""

from typing import NamedTuple
from collections import defaultdict

from ..support.foundation import transitive_closure, BreadthFirstTraversal
from .context_free import ContextFreeGrammar
from .automata import HFA, LR0_State

class ParseItemMap(NamedTuple):
	"""
	The key to the whole LR genera is the notion of a parse-item (and subsets of them).
	Normally, we think of a parse-item as a pair: a rule (or its ID, or its right-hand side)
	crossed with an index into (or just past) that right-hand side. And this is well-enough.
	However, it's an ungainly structure. I propose an alternative based on sentinels,
	so that a small integer refers to a parse-item, and the next higher integer is the
	successor parse-item. We index an array (symbol_at) to find the symbol at the dot,
	or None for the end of a rule. Other mappings facilitate each step in the dance.
	
	There's a subtle idea lurking in here:
	This structure is an interesting representation of all the rules in a context-free grammar.
	With it, the fact that parse-items are also just numbers is ALMOST completely hidden.
	"""
	symbol_at : list[str]
	rule_found : dict[int, int]
	symbol_front : dict[str: list[int]]
	language_front : list[int]
	nr_rules : int
	
	@staticmethod
	def from_grammar(grammar: ContextFreeGrammar):
		
		symbol_at, rule_found, language_front = [], {}, []
		symbol_front = {N: [] for N in grammar.symbol_rule_ids}
		
		def plonk(where, rhs):
			where.append(len(symbol_at))
			symbol_at.extend(rhs)
			rule_found[len(symbol_at)] = len(rule_found)
			symbol_at.append(None)
		
		for rule in grammar.rules:
			plonk(symbol_front[rule.lhs], rule.rhs)
		
		for symbol in grammar.start:
			plonk(language_front, [symbol])
		
		return ParseItemMap(symbol_at, rule_found, symbol_front, language_front, len(grammar.rules))


def lr0_construction(pim:ParseItemMap) -> HFA[LR0_State]:
	"""
	In broad strokes, this is a subset-construction with a sophisticated means
	to identify successor-states. The keys (by which nodes are identified) are
	core-sets of LR0 parse items. (See also _prepare_parse_items.)

	Additionally, during the full-elaboration step in visiting a core, completed
	parse-items correspond to a state's `reduce` entries. We don't worry about
	look-ahead in this construction: hence the '0' in LR(0). The net result is
	a compact table generated very quickly, but with somewhat limited power.
	In practical systems, LR(0) is normally just a first step, but some few
	grammars are deterministic in LR(0).
	"""
	
	def build_state(core_item_set: frozenset):
		def visit_parse_item(item):
			next_symbol = symbol_at[item]
			if next_symbol is None:
				rule_id = rule_found[item]
				if rule_id < nr_rules:
					reduce.append(rule_id)
			else:
				shifted_cores[next_symbol].append(item + 1)
				return symbol_front.get(next_symbol)
		
		shifted_cores, reduce = defaultdict(list), []
		transitive_closure(core_item_set, visit_parse_item)
		
		shift = {
			symbol: bft.lookup(frozenset(item_set), breadcrumb=symbol)
			for symbol, item_set in shifted_cores.items()
		}
		graph.append(LR0_State(shift=shift, reduce=reduce))

	symbol_at, rule_found, symbol_front, language_front, nr_rules = pim
	
	bft = BreadthFirstTraversal()
	initial = [bft.lookup(frozenset([item])) for item in language_front]
	graph = []
	bft.execute(build_state)
	accept = [graph[qi].shift[symbol_at[item]] for item, qi in zip(language_front, initial)]
	return HFA(graph=graph, initial=initial, accept=accept, bft=bft)




"""
This module is all about easing over the process to display where things go wrong.
A tool-developer should have easy access to sophisticated error-displays.
Such should be made both likely and comfortable independent of other tools.

Error reporting is actually a topic of some complexity. Fortunately, the common cases
are reasonably straightforward and for the uncommon ones there is YAGNI (yet).

If you can localize where an error came from, you'd generally like to include some
context in the report. The usual strategy is to show the offending line, ideally
with a specific portion highlighted somehow. If you're dealing with a text console
(as many tools do) then the `illustration` function helps: Given a single line of
text and a few parameters, it makes a suitable picture.

I prefer to keep line-breaking separate from scanning and parsing, so in those areas
a simple integer position is as much location data as you get. What's a decent means
to convert that to a line and column number? Or to slice the corresponding line
from a larger text string? The SourceText handles that. It also provides a nice
`complain(...)` method, which formats a decent-looking error message to STDOUT.

There is just one complication:

Line breaks are a funny thing. Unix calls for \n. Apple prior to OSx called for \r.
CP/M and its derivatives like Windows call for \r\n, really a printer control sequence.
OS/390 (an IBM mainframe thing) tends toward character 0x85 (a high-ascii control code)
whenever it isn't thinking in EBCDIC, but who in their right mind uses EBCDIC anymore?
The Unicode line-breaking algorithm calls for no less than ELEVEN ways to break a line!

The applications I've tested treat the Unix, Apple, and DOS conventions as line-breaks
and mostly ignore the other options defined in the Unicode standard, so that's the
default behavior of the SourceText. But you can supply a mode argument to specify
different line-ending conventions. The options are given symbolically as keys in the
LINEBREAK_MODE dictionary.

So how exactly SHOULD you delimit lines? The answer, my friend, is blowin' in the wind....
"""

import bisect, re, sys
from typing import NamedTuple, Any
from enum import Enum

LINEBREAK_MODE = {
	'normal': re.compile(r'\r\n?|\n'),
	'unicode': re.compile(r'\r\n|[\x0a-\x0d\x1c-\x1e\u0085\u2028\u2029]]'),
	'unix': re.compile(r'\n'),
	'apple': re.compile(r'\r'),
	'dos': re.compile(r'\r\n'),
}

class Severity(Enum):
	NOTICE = "Notice"
	WARNING = "Warning"
	ERROR = "Error"

class Evidence(NamedTuple):
	slice:slice
	caption: str = "here"
	
	def width(self): return self.slice.stop - self.slice.start

class Issue(NamedTuple):
	"""
	Contain all the information necessary to present elements of an error, warning, notice, or whatever.
	In theory, it should be possible to organize these for sensible presentation.
	
	The notion is that any given issue could result from an interaction of causes
	in multiple different places. (For example, a mismatched function signature
	might have declaration and reference in different files.) Thus:
	
	phase: tells what portion of the interpretation process found the issue.
	severity: tells how bad the issue is.
	description: explains the issue in plain language.
	evidence: a dictionary:
		from "key" (as known to an assumed "fetch" function),
		to lists of ``Evidence`` objects relevant to that corresponding text.
	"""
	phase: str
	severity: Severity
	description: str
	evidence: dict[Any, list[Evidence]]
	
	def as_text(self, fetch):
		"""
		This will generate a not-completely-terrible error report in text-only format.
		The precise format is subject to change, but basically you should be able to
		print this on stderr and not cry yourself to sleep.
		
		:param: "fetch" must be a function which takes a key (from the evidence dictionary)
		and returns a corresponding SourceText object.
		"""
		lines = ["%s %s: %s"%(self.phase, self.severity, self.description)]
		for key, evidence in self.evidence.items():
			source = fetch(key)
			if source.filename:
				lines.append("Excerpt from "+source.filename+" :")
			for e in evidence:
				row, col = source.find_row_col(e.slice.start)
				single_line = source.line_of_text(row)
				lines.append(illustration(single_line, col, e.width(), prefix='% 6d :'%row, caption=e.caption))
		return "\n".join(lines)
		
	def emit(self, fetch):
		""" Print to standard error the generated error text. """
		print(self.as_text(fetch), file=sys.stderr)

def illustration(single_line:str, start:int, width:int=0, *, prefix='', caption="near here") -> str:
	""" Builds up a picture of where something appears in a line of text. Useful for polite error messages. """
	blanks = ''.join(c if c == '\t' else ' ' for c in prefix + single_line[:start])
	underline_width = max(1, min(width, len(single_line)-start))
	underline = '^'*underline_width
	return prefix + single_line.rstrip() + '\n' + blanks + underline +" "+caption

class SourceText:
	""" Wrapper for (a section of) source text: participates in half-respectable error-display with context. """
	def __init__(self, content:str, line_breaks='normal', filename:str=None, first_line=1):
		self.content = content
		self.filename = filename
		self.line_breaks = line_breaks
		self.first_line = first_line
	
	def __make_bounds(self):
		""" Lazily only find line breaks if it turns out to be necessary for a particular text. """
		if not hasattr(self, '__bound'):
			inside = [m.end() for m in LINEBREAK_MODE[self.line_breaks].finditer(self.content)]
			self.__bounds = [0] + inside + [len(self.content)]

	def find_row_col(self, index:int):
		""" Based on a character index offset from the start of text. Respects self.first_line. """
		self.__make_bounds()
		row = bisect.bisect_right(self.__bounds, index, hi=len(self.__bounds) - 1) - 1
		col = index - self.__bounds[row]
		return row+self.first_line, col
	
	def line_of_text(self, row):
		""" Argument respects self.first_line. """
		r = max(0, row - self.first_line)
		return self.content[self.__bounds[r]:self.__bounds[r + 1]]
	
	def complain(self, a_slice:slice, message:str):
		left, right = a_slice.start, a_slice.stop
		row, col = self.find_row_col(left)
		line = self.line_of_text(row)
		if self.filename is None: print("At line %d, column %d:" % (row, col + 1), message, file=sys.stderr)
		else: print("%s:%d column %d: " % (self.filename, row, col + 1), message, file=sys.stderr)
		print(illustration(line, col, right-left, prefix=' >>> '), file=sys.stderr)



"""
"""
from .interface import INITIAL, END_OF_INPUT, FiniteAutomaton, Bindings, RuleId

def _make_reader(subject):
	if isinstance(subject, str):
		return lambda offset: ord(subject[offset])
	elif isinstance(subject, (bytes, bytearray)):
		return subject.__getitem__
	else:
		raise TypeError()

def scan_one_raw_lexeme(fa:FiniteAutomaton, codepoint_at, cursor, condition:str) -> tuple[int, RuleId]:
	"""
	This is how the magic happens. Given a start-condition, this algorithm finds the
	end of the lexeme to be matched and the rule number (if any) that applies to the match.
	This algorithm deliberately ignores a zero-width, zero-context match, but it may fail
	to consume any characters. The caller must deal properly with that case.
	"""
	right, rule_id, jammed = cursor, None, fa.jam_state()
	is_at_start_of_line = cursor == 0 or codepoint_at(cursor - 1) in {10, 13}
	state = fa.condition(condition)[is_at_start_of_line]
	while True:
		try: codepoint = codepoint_at(cursor)
		except IndexError: codepoint = END_OF_INPUT
		state = fa.transition(state, codepoint)
		if state == jammed: break
		cursor += 1
		accept = fa.accept(state)
		if accept is not None:
			right, rule_id = cursor, accept
	return right, rule_id

class Scanner:
	"""
	This is the standard generic finite-automaton-based scanner, with support for backtracking,
	beginning-of-line anchors, and (non-variable) trailing context.
	"""
	
	condition : str
	
	def __init__(self, text, fa:FiniteAutomaton, bindings:Bindings, start=INITIAL, at=0):
		self.__text = text
		self.__size = len(text)
		self.__read = _make_reader(text)
		self.__fa = fa
		self.__bindings = bindings
		self.__stack = []
		self.enter(start)
		self.left = self.right = at
		
	def scan_one_item(self):
		cursor = self.left = self.right
		self.right, rule_id = scan_one_raw_lexeme(self.__fa, self.__read, cursor, self.condition)
		if rule_id is None:
			self.right = cursor + 1
			self.__bindings.on_stuck(self)
		else:
			self.__bindings.on_match(self, rule_id)
	
	def scan_repeatedly(self):
		while self.has_more():
			self.scan_one_item()
	
	def has_more(self):
		return self.right < self.__size
	
	def enter(self, condition):
		""" Change to a different start-condition. """
		self.condition = condition
	
	def push(self, condition):
		""" Push the current start-condition onto a stack and change to the one specified. """
		self.__stack.append(self.condition)
		self.enter(condition)
	
	def pop(self):
		""" Jump back to a start-condition pulled from the condition stack. """
		self.enter(self.__stack.pop())
		
	def slice(self):
		""" Return a slice-object corresponding to the extent of matched text. """
		return slice(self.left, self.right)
	def match(self):
		""" Return the actual matched text """
		return self.__text[self.left:self.right]

	def less(self, nr_chars):
		"""
		Put trailing characters back into the stream to be matched.
		This is also the mechanism for trailing context.
		if `nr_chars` is zero or positive, the current match is adjusted to
		consume precisely that many characters. If it's negative, then a
		corresponding number of characters are dropped from the end of the
		match, and these will be considered again in the next matching cycle.
		(If `nr_chars` is None, nothing happens.)
		"""
		if nr_chars is not None:
			self.right = (self.right if nr_chars < 0 else self.left) + nr_chars
	
	def seek(self, position):
		""" Scanning will next resume at the position given. """
		self.right = position

class IterableScanner(Scanner):
	"""
	It may be convenient that iterating over a scanner would cause it to yield tokens.
	Scan-actions can call yy.token(...) and this object wraps that into an iterable.
	"""
	
	def __init__(self, text, fa: FiniteAutomaton, bindings: Bindings, start=INITIAL, at=0):
		super().__init__(text, fa, bindings, start, at)
		self.__buffer = []
	
	def __iter__(self):
		while self.has_more():
			self.scan_one_item()
			yield from self.__buffer
			self.__buffer.clear()
	
	def token(self, kind: str, semantic=None):
		"""
		During scan rule invocation, call this method with tokens for the
		scanner to yield once it gets control back. For integration with
		the parsing mechanism, a token is defined as a 2-tuple consisting
		of "kind" and arbitrary/optional semantic value.
		"""
		assert kind is not None
		self.__buffer.append((kind, semantic))



import setuptools

setuptools.setup(
	name='booze-tools',
	version='0.6.0.2',
	packages=[
		'boozetools',
		'boozetools.macroparse',
		'boozetools.parsing',
		'boozetools.parsing.general',
		'boozetools.scanning',
		'boozetools.support',
	],
	description='A panoply of tools for parsing, lexical analysis, and semantic processing',
	long_description=open('README.md').read(),
	long_description_content_type="text/markdown",
	url="https://github.com/kjosib/booze-tools",
	classifiers=[
		"Programming Language :: Python :: 3.9",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Topic :: Software Development :: Compilers",
		"Development Status :: 3 - Alpha",
    ],
	author="Ian Kjos",
	author_email="kjosib@gmail.com"
)


"""
Dealing with anything like a programming language, you need a symbol table.
Here's my concept:

I distinguish between:

* name: a distinct sequence of characters with the appearance of a name.
* symbol: an association between name, declaration, definition, namespace, and scope. This association may be a work in progress during certain phases of translation.
* alias: another name for the same symbol, possibly used in a different scope or namespace. (By implication, symbols should record their original names.)
* declaration: something in the text that lets you know a name will be valid, and perhaps a few attributes of its meaning (within some namespace and scope).
* definition: something in the text that tells you exactly what a name means (within a namespace and scope). Normally, whatever you define you implicitly also declare.
* mention: the act of using a name to refer to whatever it signifies, or the place where this is done. In some languages, mentioning a name implicitly declares a corresponding symbol.
* namespace: a particular group of names with similar rules for interpreting them. Names that look the same, but which appear in different namespaces, do not refer to the same thing. Generally something about the local syntax tells you what namespace applies to a name. (For example, in SQL, tables and columns are in separate namespaces.)
* scope: a particular section of text, or of the semantic structure associated with that text. We often think of scope as contiguous and nested, but record-field access is a counterexample. There can be more than one namespace associated with a scope.
* context: the combination of namespace and scope in which some name appears within the text.
* symbol table: the organized collection of all symbols defined or used in the text. The translator may consult or modify the symbol table as needed.
* environment: the slice of the symbol table (and extra related information) in view while translating a particular segment of text.

In this module, I propose a generic structure for symbols, namespaces, and maybe a few other bits.
With luck, it should be suitable for both translation and reasonable error reporting.
One obvious failing is its inherently imperative nature: It relies on mutation.
For the moment, I'll live with that.

I am not concerned with distinguishing between a static and dynamic environment.
Rather, I assume that "the environment" is whatever your application needs for its own interpretive dance.
Thus, the precise structure of "the environment" is not defined here. Your application should define a suitable one.
However, the usual notion is probably that of one or more name-spaces in which the translator might search for
or add names and their definitions.

There are also a few typical patterns of interaction with the environment (and by extension, with namespaces).
The most common are probably:

* adding a name, expecting it not already to exist locally (or at all, if you prohibit shadowing).
* looking up a name, expecting it surely to have an already-associated symbol.
* looking up a name, with the plan to add a fresh symbol locally if one is not found.

The key question is what comes next when an expectation is falsified.
For now, I'll just go with raising an exception.
Why? Because it's Python.

Anyway, you'll probably represent "the environment" as one or more namespaces and maybe some other attributes.
"""

from typing import Optional, Generic, TypeVar

class NoSuchSymbol(KeyError):
	pass

class SymbolAlreadyExists(KeyError):
	pass

T = TypeVar("T")

class NameSpace(Generic[T]):
	"""
	NameSpace bears some resemblance to chainmap with a few extra attributes.
	The "local" is the set of words/names defined in this space.
	The "place" is a general statement of where the namespace "lives", used for error messages.
	The "parent" works like a static link.
	
	What's particularly noteworthy here is the application of static type hinting.
	"""
	def __init__(self, *, place, parent:Optional["NameSpace[T]"]=None):
		self.local : dict[object:T] = {}
		self.place = place
		self.parent : NameSpace[T] = parent
		
	def find(self, key) -> tuple[Optional[T], Optional["NameSpace[T]"]]:
		"""
		Frequently when dealing with nested namespaces in translation,
		you need to know in which layer of the nesting structure a name appears.
		This method usefully returns both the symbol and its host namespace.
		Assuming you've decorated namespace objects to your needs,
		(perhaps via the "place" field)
		you should have no trouble working out e.g. which stack frame you find a variable in.
		In case the symbol is not found, this will return (None, None) for easy procedural testing.
		Why? Because I've
		"""
		if key in self.local:
			return self.local[key], self
		elif self.parent is not None:
			return self.parent.find(key)
		else:
			raise NoSuchSymbol(key)
	
	def replace(self, key, value:T):
		""" Suppose you need to replace a symbol. Fine. But you're going to know it. """
		if key in self.local:
			self.local[key] = value
		else:
			raise NoSuchSymbol(key)
	
	def __getitem__(self, key):
		if key in self.local:
			return self.local[key]
		elif self.parent is not None:
			return self.parent[key]
		else:
			raise NoSuchSymbol(key)
		
	def __contains__(self, key):
		return key in self.local or (self.parent is not None and key in self.parent)
	
	def __setitem__(self, key, value:T):
		if key in self.local:
			raise SymbolAlreadyExists(key)
		else:
			self.local[key] = value

	def new_child(self, place) -> "NameSpace[T]":
		""" Return a subordinate name-space linked to this one. """
		return NameSpace(place=place, parent=self)
